//-----------------------------------------------
// MIDI-CV Interface for Gakken Analog Synth SX-150
// Author   : RJB
// Device   : ATtiny2313
// X'tal    : 20MHz
// FUSE     : -FL11101111
// Version  : 0.1alfa May. 29 2008
//-----------------------------------------------
// include

#include <avr\io.h>
#include <avr\signal.h>
#include <avr\interrupt.h>

//-----------------------------------------------
// define

#define GATE1		PD6	

#define GATE1_ON		(PORTD &= ~(1 << GATE1))
#define GATE1_OFF		(PORTD |= (1 << GATE1))

#define	TRUE			1
#define	FALSE			0
#define ON				1
#define OFF				0

#define MAX_NOTE_CNT	8
#define	BUFSIZE			32		// Oobt@TCY
#define	INTRCOUNT		178		// ^C} 250us

// MIDIbZ[W
#define MIDI_NoteOff			0x80
#define MIDI_NoteOn				0x90
#define MIDI_PolykeyPressure	0xA0
#define MIDI_ControlChange		0xB0
#define MIDI_ProgramChange		0xC0
#define MIDI_ChannelPressure	0xD0
#define MIDI_PitchBend			0xE0

#define MIDI_StartSysEx			0xF0
#define MIDI_EndSysEx			0xF7

#define MIDI_TimingClock		0xF8
#define MIDI_Start				0xFA
#define MIDI_Continue			0xFB
#define MIDI_Stop				0xFC
#define MIDI_ActiveSensing		0xFE
#define MIDI_SystemReset		0xFF

//-----------------------------------------------
// f[^^`

typedef char			schar;
typedef unsigned char	uchar;
typedef	int				int16;
typedef	unsigned int	uint16;
typedef	long			int32;
typedef	unsigned long	uint32;
typedef	unsigned char	BOOL;

//-----------------------------------------------
// O[oϐ

uchar	RxByte;
uchar	PC;
uchar	SysEx;
uchar	Rch;
uchar	MByte[2];
uchar	ch_bit;
uchar	gate;
uchar	trig;

uchar	PBBuf;
uchar	NoteCnt;
uchar	NoteBuf[MAX_NOTE_CNT];
uchar	PlayNoteBuf;
uint16  CVduty;

// Oobt@p
uchar	RxBuf[BUFSIZE];
int16	Ptr_buf_in, Ptr_buf_out; // |C^

//-----------------------------------------------
// USART荞

SIGNAL(SIG_USART0_RX)
{
	// Oobt@֕ۑ
	Ptr_buf_in++;
	Ptr_buf_in &= (BUFSIZE - 1);
	RxBuf[Ptr_buf_in] = UDR;
}

//-----------------------------------------------
// ^C}0I[o[t[荞

SIGNAL(SIG_TIMER0_OVF)
{
	if(trig) {
		OCR1A = 1023;
		trig = 0;				
	}
	else {
		OCR1A = gate ? CVduty : 0;
	}
}

//-----------------------------------------------
// CVo͓dZbg

void SetCV(void)
{
	uchar note;
	uchar m;

	note = PlayNoteBuf;
	if( note < 24) 	note = 24;
	if( note > 84) 	note = 84;

	m = ((note - 24) * 3) / 4;

	CVduty = note * 8 + (PBBuf / 4) + 248 + m;
}

//-----------------------------------------------
// m[gON

void NoteON(void)
{
	uchar i;
	uchar max = 0;

	// obt@ɓo^ς̃m[g͖
	for(i = 0; i < NoteCnt; i++) {
		if(NoteBuf[i] == MByte[0]) {
			return;
		}
	}

	// obt@ςH
	if(NoteCnt == MAX_NOTE_CNT) {
		// ʓ˂
		for(i = 0; i < (MAX_NOTE_CNT - 1); i++) {
			NoteBuf[i] = NoteBuf[i+1];
		}
		NoteBuf[MAX_NOTE_CNT - 1] = MByte[0];
	}
	else {
		NoteBuf[NoteCnt] = MByte[0];
		NoteCnt++;
	}

	// ō
	for(i = 0; i < NoteCnt; i++) {
		if(max < NoteBuf[i]) {
			max = NoteBuf[i];
		}
	}
	PlayNoteBuf = max;

	SetCV();
	GATE1_ON;
	gate = 1;
	trig = 1;
}

//-----------------------------------------------
// m[gOFF

void NoteOFF(void)
{
	uchar i;
	uchar max = 0;
	BOOL flg = FALSE;

	// obt@ɓo^Ăꍇ͍폜
	for(i = 0; i < NoteCnt; i++) {
		if(flg) {
			NoteBuf[i-1] = NoteBuf[i];
		}
		if(NoteBuf[i] == MByte[0]) {
			flg = TRUE;
		}
	}
	if(flg) NoteCnt--;

	if(NoteCnt == 0) {
		// obt@J̏ꍇGATEIt
		GATE1_OFF;
		gate = 0;
	}
	else {
		// ō
		for(i = 0; i < NoteCnt; i++) {
			if(max < NoteBuf[i]) {
				max = NoteBuf[i];
			}
		}
		PlayNoteBuf = max;
		SetCV();
	}
}

//-----------------------------------------------
// MIDIVXeA^CbZ[W

void MIDI_SystemRealtimeMessage(void)
{
	switch(RxByte) {
		case MIDI_TimingClock:
			break;
		case MIDI_Start:
			break;
		case MIDI_Continue:
			break;
		case MIDI_Stop:
			break;
		case MIDI_ActiveSensing:
			break;
		case MIDI_SystemReset:
			break;
	}
}

//-----------------------------------------------
// MIDIVXebZ[W

BOOL MIDI_SystemMessage(void)
{
	if(SysEx){
		if(RxByte == MIDI_EndSysEx){
			SysEx = FALSE;
		}
	}
	else{
		if(RxByte < 0xF8) {
			if(RxByte > 0x7F){
				if(RxByte == MIDI_StartSysEx){
					SysEx = TRUE;
				}
				else{
					Rch = RxByte & 0x0F;
               	}
				PC = 0;
 			}
			else{
				MByte[PC & 0x01] = RxByte;
			}
			return TRUE;
		}
		else {
			MIDI_SystemRealtimeMessage();
		}
	}
	return FALSE;
}

//-----------------------------------------------
// MIDI`lbZ[W

void MIDI_ChannelMessage(void)
{
	switch(PC){
		case 0:
			switch(RxByte & 0xF0){
			case MIDI_NoteOff:
				PC = 2;
				break;
			case MIDI_NoteOn:
				PC = 4;
				break;
			case MIDI_PolykeyPressure:
				PC = 6;
				break;
			case MIDI_ProgramChange:
				PC = 8;
				break;
			case MIDI_ControlChange:
				PC = 10;
				break;
			case MIDI_ChannelPressure:
				PC = 12;
				break;
			case MIDI_PitchBend:
				PC = 14;
				break;
			} break;

        // m[gIt
		case 2:
			PC = 3;
			break;
        case 3:
			PC = 2;
			NoteOFF();
			break;

        // m[gI
        case 4:
			PC = 5;
			break;
        case 5:
			PC = 4;
			if( MByte[1] == 0){
				NoteOFF();
			}
			else{
				NoteON();
			}
			break;

		// |tHjbNL[vbV[ 
        case 6:
			PC = 7;
			break;
        case 7:
			break;

        // vO`FW
        case 8:
 			break;

        // Rg[`FW
        case 10:
			 PC = 11;
			 break;
        case 11:
			break;

        // `lvbV[
		case 12:
			PC = 13;
			break;
		case 13:
			break;

        // sb`xh
		case 14:
			PC = 15;
			break;
		case 15:
			PBBuf = MByte[1];
			SetCV();
			break;

		default:
			break;
    }
}

//-----------------------------------------------
// C

int16 main(void)
{
	DDRB = 0xFF;
	DDRD = 0xFE;

	PBBuf = 64;
	NoteCnt = 0;
	CVduty = 0;
	gate = 0;
	trig = 0;

	// USART CLOCK=20MHz BAUD=31.25kHz
	// UBRR = (CLOCK / (16BAUD)) - 1 = 39
	UCSRB = 0x00;
	UCSRA = 0x00;
	UCSRC = 0x06;	// 8bit STOP1bit
	UBRRL = 0x27;	// 39
	UBRRH = 0x00;	
	UCSRB |= (1 << RXCIE) | (1 << RXEN);
	Ptr_buf_in = Ptr_buf_out = 0;

	// ^C}0ݒ(荞)
	TCCR0A = 0x00;
	TCCR0B = 0x03; // 1/64Clk
	TIFR |= (1 << TOV0);
	TIMSK |= (1<< TOIE0);
	TCNT0 = INTRCOUNT;

	// ^C}1ݒ(PWM)
	TCNT1 = 0x0000;
	TCCR1B = 0x00;
	TCCR1A = 0xA3;
    OCR1A = 0x0000;	// CV 
    OCR1B = 0x0000;
	TCCR1B = 0x09;

	// 荞݋
	sei();

	// C[v
	while(TRUE) {
		// Oobt@H
		if(Ptr_buf_in == Ptr_buf_out) 	continue;

		// Oobt@PoCgo
		Ptr_buf_out++;
		Ptr_buf_out &= (BUFSIZE - 1);
		RxByte = RxBuf[Ptr_buf_out];

		if(MIDI_SystemMessage()) {
			if(Rch == 0) {
				MIDI_ChannelMessage();
			}
		}
	}
	return(0);
}
